/*!
\file init.c
\author Robert Sperling <sperling@small-window-manager.de>
\version 1.3.6
\brief everything needed to initialize sWM

\verbatim
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
\endverbatim
*/

#ifndef PIXMAPDIR
//! location of directory pixmaps containing a valid "theme"
#define PIXMAPDIR "."
#endif//PIXMAPDIR

/***************************************************************************/

// -scanCurrentWindows--------------------------------------
//! puts all existing windows into the database of sWM 
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre3: got invalid windows?!?! ignoring them
   v1.3.0-pre5: added grabServer
   v1.3.6:      now ignoring sWM-Window
\endverbatim
*/
void scanCurrentWindows()
{
  unsigned int counter, numberOfWindows;
  Window dw1, dw2, *windowList;
  XWindowAttributes attr;
#ifdef TRACE
  trace(1,"scanCurrentWindows");
#endif//TRACE

#ifdef CENTERNEW
  scanWindowsRunning=true;
#endif//CENTERNEW
  
#ifdef VERBOSE
  swmerr ("querying x-window tree");
#endif//VERBOSE

  // Needed to ensure that no window dies before we are ready
  XGrabServer (display);

  XQueryTree(display, rootWindow, &dw1, &dw2, &windowList, &numberOfWindows);
  
  for (counter = 0; counter < numberOfWindows; counter++) 
  { XGetWindowAttributes(display, windowList[counter], &attr);
#ifdef GNOME
    if (windowList[counter] && windowList[counter]!=sWMWindow)
#else//GNOME
    if (windowList[counter])
#endif//GNOME
     if (!attr.override_redirect && attr.map_state == IsViewable)
     { createNewClient(windowList[counter]);
     }
  }// end for
  
#ifdef VERBOSE
  swmerr ("Xfreeing memory in scanCurrentWindows");

#ifdef CHECKCLIENTS
 dumpClients();
#endif//CHECKCLIENTS
#endif//VERBOSE
  if (windowList) XFree(windowList);
  
  XUngrabServer (display);

#ifdef CENTERNEW
  scanWindowsRunning=false;
#endif//CENTERNEW

#ifdef TRACE
  trace(-1,"scanCurrentWindows");
#endif//TRACE
}//scanCurrentWindows 

// -initSWM----------------------------------------------------
//! does exactly what it is called like, initializing sWM
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: no change
   v1.3.0-pre3: changed font errors to fprintf
   v1.3.0-pre4: removes font stuff from where it is not needed
   v1.3.5:	invertGC is now in minimal mode only
   v1.3.6:      added color for sWM-Panel
\endverbatim
*/
   
void initSWM()
{ XGCValues 		gv;
  XSetWindowAttributes 	attr;
  XColor 		dummy;
#ifndef NOTHEMES
  char *		env;
#endif//NOTHEMES

#ifdef TRACE
  trace(1,"initSWM");
#endif//TRACE

#ifdef VERBOSE
  swmerr ("setting signals & error handler");
#endif//VERBOSE

  XSetErrorHandler(xErrorHandler);

#ifdef SIGNALS
  signal(SIGTERM, signalHandler);
  signal(SIGINT,  signalHandler);
  signal(SIGHUP,  signalHandler);
#ifdef DEBUG
  signal(SIGUSR1, signalHandler);
  signal(SIGUSR2, signalHandler);
#endif//DEBUG
#endif//SIGNALS

#ifdef VERBOSE
  swmerr ("opening display");
#endif//VERBOSE

  // opens display given in args 
  display = XOpenDisplay(displayOption); 

  if (!display) 
  { 
#ifdef STD_IO  
    fprintf(stdout,"sWM: cannot open display %s, giving up\n", displayOption);
#endif//STD_IO
    exit(1);
  }// end if no display 
  screen = DefaultScreen(display);
  rootWindow = RootWindow(display, screen);

#ifdef MAXIMIZE  
  xmax = DisplayWidth (display, screen);
  ymax = DisplayHeight(display, screen);
#else//MAXIMIZE
#ifdef CENTERNEW
  xmax = DisplayWidth (display, screen);
  ymax = DisplayHeight(display, screen);
#endif//CENTERNEW
#endif//MAXIMIZE

  
#ifdef VERBOSE
  swmerr ("allocating colors");
#endif//VERBOSE

  // colors used by SWM itself 

#ifndef MINIMAL
#ifndef NOTHEMES
  env=getenv("SWM_FOREGROUND_COLOR");
  if (env)
  {
#ifdef VERBOSE  
    swmerr ("got SWM_FOREGROUND_COLOR, it will be: %s",env);
#endif//VERBOSE    
    XAllocNamedColor(display, DefaultColormap(display, screen), 
                      env, &foregroundColor, &dummy);
  }	      
  else
#endif//NOTHEMES
  { 
#ifdef VERBOSE  
    swmerr ("no sWM_FOREGROUND_COLOR found, using default");
#endif//VERBOSE
    XAllocNamedColor(display, DefaultColormap(display, screen), 
                     givenFG, &foregroundColor, &dummy);
  }		     
#endif//MINIMAL

#ifndef NOTHEMES
  env=getenv("SWM_BACKGROUND_COLOR");
  if (env)
  { XAllocNamedColor(display, DefaultColormap(display, screen), 
                     env, &backgroundColor, &dummy);
  }	      
  else
#endif//NOTHEMES
    XAllocNamedColor(display, DefaultColormap(display, screen), 
                     givenBG, &backgroundColor, &dummy);

#ifndef NOTHEMES
  env=getenv("SWM_BORDER_COLOR");
  if (env)
  { XAllocNamedColor(display, DefaultColormap(display, screen), 
                     env, &borderColor, &dummy);
  }	      
  else
#endif//NOTHEMES
    XAllocNamedColor(display, DefaultColormap(display, screen), 
                     givenBC, &borderColor, &dummy);

#ifdef SWMPANEL
#ifndef NOTHEMES
  env=getenv("SWM_PANEL_COLOR");
  if (env)
  { XAllocNamedColor(display, DefaultColormap(display, screen), 
                     env, &sWMPanelColor, &dummy);
  }	      
  else
#endif//NOTHEMES
    XAllocNamedColor(display, DefaultColormap(display, screen), 
                     givenPC, &sWMPanelColor, &dummy);

#ifndef NOTHEMES
  env=getenv("SWM_PANELTEXT_COLOR");
  if (env)
  { XAllocNamedColor(display, DefaultColormap(display, screen), 
                     env, &sWMPanelTextColor, &dummy);
  }	      
  else
#endif//NOTHEMES
    XAllocNamedColor(display, DefaultColormap(display, screen), 
                     givenPTC, &sWMPanelTextColor, &dummy);
#endif//SWMPANEL

#ifdef SHOWFOCUS
  XAllocNamedColor(display, DefaultColormap(display, screen), 
                   givenActiveBC, &activeBorderColor, &dummy);
#endif//SHOWFOCUS

#ifdef FONTS
  fontAvailable = true;
  sWMFont = XLoadQueryFont(display, givenFont);

  if (sWMFont==(XFontStruct *) NULL)
  { /* the given Font */
#ifdef ERRORMSG
    fprintf(stderr,"sorry, can't load font %s\n", givenFont);
#endif//ERRORMSG

    sWMFont = XLoadQueryFont(display, fallbackFont);
    if (sWMFont==(XFontStruct *) NULL)
    { // fallback font not available 
#ifdef ERRORMSG
      fprintf(stderr,"sorry, can't even load font %s\n",fallbackFont);
#endif//ERRORMSG    
      sWMFont = XLoadQueryFont(display, "fixed");
      if (sWMFont==(XFontStruct *) NULL)
      { // even fixed font not available 
#ifdef ERRORMSG
        fprintf(stderr,"sorry, can't even load font fixed, disabling fonts\n");
#endif//ERRORMSG
        fontAvailable = false;
      }// end if no fixed
    }// end loading fixed
  }// end loading fallback font
#endif//FONTS

#ifdef CURSORS
#ifdef VERBOSE
  swmerr ("creating cursors");
#endif//VERBOSE

  moveCursor  	= XCreateFontCursor(display, MOVECURSOR);
  resizeCursor 	= XCreateFontCursor(display, RESIZECURSOR);
  bodyCursor    = XCreateFontCursor(display, BODYCURSOR);
  windowCursor  = XCreateFontCursor(display, WINDOWCURSOR);
  // XC_X_cursor
#endif//CURSORS

  gv.function  	= GXcopy;
#ifndef MINIMAL
  gv.foreground	= foregroundColor.pixel;
#else//MINIMAL
  gv.foreground	= borderColor.pixel;
#endif//MINIMAL
  
#ifdef FONTS
  gv.font 	= sWMFont->fid;

  stringGC 	= XCreateGC (display, rootWindow, 
                             GCFunction | GCForeground | GCFont, &gv);
#endif//FONTS

  gv.foreground = borderColor.pixel;
  gv.line_width = 1;
  
  borderGC 	= XCreateGC (display, rootWindow, 
                             GCFunction | GCForeground | GCLineWidth, &gv);

#ifdef FAST
// TESTING: GCFont was in parameters (?!?)
  gv.function 	= GXinvert;
  gv.subwindow_mode = IncludeInferiors;
  invertGC 	= XCreateGC (display, rootWindow, 
  		             GCFunction | GCSubwindowMode |  
			     GCLineWidth | GCForeground, &gv);
#endif//FAST 

#ifdef SWMPANEL
  gv.foreground = sWMPanelTextColor.pixel;
  gv.background = sWMPanelColor.pixel;
  panelGC	= XCreateGC (display, rootWindow, 
  		             GCFunction | GCBackground | 
			     GCLineWidth | GCForeground | GCFont, &gv);
#endif//SWMPANEL


#ifdef LOWCOLOR
  attr.event_mask = ExposureMask | ChildMask | ColormapChangeMask | 
                    ButtonMask | PropertyChangeMask;
#else
//TESTING: no ExposureMask
  attr.event_mask = ChildMask | ButtonMask | 
                    PropertyChangeMask;
#endif//LOWCOLOR

#ifdef VERBOSE
  swmerr ("changing windowsAttributes");
#endif//VERBOSE

  XChangeWindowAttributes(display, rootWindow, CWEventMask, &attr);

  xa_wm_state        = XInternAtom(display, "WM_STATE", False);
#ifdef MWM
  xa_mwm_wm_hints    = XInternAtom(display, "_MOTIF_WM_HINTS", False);
#endif//MWM
  xa_wm_change_state = XInternAtom(display, "WM_CHANGE_STATE", False);
  xa_wm_protos       = XInternAtom(display, "WM_PROTOCOLS", False);
  xa_wm_delete       = XInternAtom(display, "WM_DELETE_WINDOW", False);

#ifdef LOWCOLOR
  xa_wm_cmapwins     = XInternAtom(display, "WM_COLORMAP_WINDOWS", False);
#endif//LOWCOLOR

#ifdef LAYERS
  xa_windowLayer     = XInternAtom(display, "_WIN_LAYER", False);
#endif//LAYERS

#ifdef KEYS
 grabKeys();
#endif//KEYS

#ifdef PIXMAPS
 initPixmaps();
#endif//PIXMAPS
 
#ifdef TRACE
  trace(-1,"initSWM");
#endif//TRACE
}//initSWM 

/**************************************************************************/

#ifdef KEYS
//  grabKeys
//! grabs keys only to root window
/*!
\verbatim
   v1.2.0:      initial release
   v1.3.0-pre1: no changes
   v1.3.0-pre2: no changes
   v1.3.0-pre3: no changes
\endverbatim
*/

void grabKeys() 
{  
#ifdef TRACE 
  trace(1,"grabKeys");
#endif//TRACE

  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(CYCLEKEY)),
           Mod1Mask,rootWindow,True,GrabModeAsync,GrabModeAsync);

#ifdef VDESKTOPS
#ifdef VERBOSE
  swmerr ("grabbing vDesktop keys");
#endif//VERBOSE  
  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(CYCLEKEY)),
           ControlMask,rootWindow,True,GrabModeAsync,GrabModeAsync);
#endif//VDESKTOPS

  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(CLOSEKEY)),
           Mod1Mask,rootWindow,True,GrabModeAsync,GrabModeAsync);

#ifdef KEYRAISE
  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(RAISEKEY)),
           Mod1Mask,rootWindow,True,GrabModeAsync,GrabModeAsync);
#endif//KEYRAISE

#ifdef MAXIMIZE
#ifdef KEYMAXIMIZE
  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(MAXIMIZEKEY)),
           Mod1Mask,rootWindow,True,GrabModeAsync,GrabModeAsync);
#endif//KEYMAXIMIZE
#endif//MAXIMIZE

#ifdef LAYERS
  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(LAYERUP)),
           Mod1Mask,rootWindow,True,GrabModeAsync,GrabModeAsync);
  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(LAYERDOWN)),
           Mod1Mask,rootWindow,True,GrabModeAsync,GrabModeAsync);
#endif//LAYERS

#ifdef MWM  
  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(FRAMEKEY)),
           Mod1Mask,rootWindow,True,GrabModeAsync,GrabModeAsync);
#endif//MWM

#ifdef DEBUG    
  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(DEBUGKEY)),
           Mod1Mask,rootWindow,True,GrabModeAsync,GrabModeAsync);
#endif//DEBUG

#ifdef EXITKEY
  XGrabKey(display,XKeysymToKeycode(display,XStringToKeysym(SWMEXITKEY)),
           Mod1Mask,rootWindow,True,GrabModeAsync,GrabModeAsync);
#endif//EXITKEY


#ifdef TRACE 
  trace(-1,"grabKeys");
#endif//TRACE

}// end grabKeys
#endif//KEYS



#ifdef PIXMAPS
// -readPixmap-------------------------------------------------
//! loads a pixmap from a file
/*!
\param fileName		the file to load into pm
\param pm		data structure to hold loaded picture

\verbatim
   v1.3.0-pre2: initial release
   v1.3.0-pre3: no change
   v1.3.6:      wrapped errormessages into ERRORMSG
\endverbatim
*/ 
void readPixmap(char *fileName, pixmapStruct *pm) 
{ XpmAttributes attr; 
  int		val;
 
#ifdef TRACE
  trace(1,"readPixmap"); 
#endif//TRACE
 
  attr.valuemask=XpmSize|XpmColorSymbols; 
  attr.colorsymbols=NULL; 
  attr.numsymbols=0; 
 
  // Read pixmap and get size 
  if((val=XpmReadFileToPixmap(display,rootWindow,fileName, 
                              &(pm->pixmap),&(pm->mask), &attr))==XpmSuccess) 
  { pm->width=attr.width; 
    pm->height=attr.height; 
#ifdef VERBOSE
    swmerr ("pixmap read");
#endif//VERBOSE
  }
  else
  { 
#ifdef ERRORMSG  
    if(val == XpmOpenFailed)                                                  
    {
#ifdef STD_IO
     fprintf(stderr, "XPM: Couldn't open pixmap at: %s\n",fileName);
#endif//STD_IO
    }     
    else if(val == XpmFileInvalid)                                            
#ifdef STD_IO
     fprintf(stderr, "XPM: Invalid pixmap format\n");
#endif//STD_IO     
    else if(val == XpmColorError)                              
#ifdef STD_IO
     fprintf(stderr, "XPM: Invalid color information in file\n");               
#endif//STD_IO
    else if(val == XpmNoMemory)                                               
#ifdef STD_IO
     fprintf(stderr, "XPM: Could not allocate memory\n");
#endif//STD_IO     
    else if(val == XpmColorFailed)                                            
#ifdef STD_IO
     fprintf(stderr, "XPM: could not allocate colors\n");
#endif//STD_IO     
#endif//ERRORMSG
    exit(1);                                                                  
  }// end else if XpmSuccess 
#ifdef TRACE
  trace(-1,"readPixmap");
#endif//TRACE
}//readPixmap 

/**************************************************************************/

//  initPixmaps
//! loads all required pixmap files
/*!
\verbatim
   v1.3.0-pre2: initial release
   v1.3.0-pre3: no change
   v1.3.1:      now pixmapdir and theme can be customized
\endverbatim   
*/
 
void initPixmaps()
{
#ifdef SWMPANEL
  char toLoad[255];
  int  counter;
#endif//SWMPANEL
  
#ifdef TRACE 
  trace(1,"initPixmaps");
#endif//TRACE

 readPixmap(PIXMAPDIR"/theme/background.xpm", &backgroundPix); 
 readPixmap(PIXMAPDIR"/theme/closebox.xpm", &closeBoxPix); 
 readPixmap(PIXMAPDIR"/theme/maxbox.xpm", &maxBoxPix); 
 readPixmap(PIXMAPDIR"/theme/stickybox.xpm", &stickyBoxPix); 
 readPixmap(PIXMAPDIR"/theme/unstickybox.xpm", &unStickyBoxPix); 
#ifdef SWMPANEL
 for (counter=0; counter<NUMBEROFSTARTERS; counter++)
 { sprintf(toLoad,"%s/starters/starter%i.xpm",PIXMAPDIR,counter+1);
#ifdef VERBOSE
   fprintf(stderr,toLoad);
#endif//VERBOSE   
   readPixmap(toLoad, &starterPix[counter]); 
 }
#endif//SWMPANEL

#ifdef TRACE 
  trace(-1,"initPixmaps");
#endif//TRACE
}//initPixmap 

/**************************************************************************/

//  releasePixmaps
//! frees all allocated pixmap files
/*!
\verbatim
   v1.3.0-pre4: initial release
\endverbatim   
*/ 
void releasePixmaps()
{ 
#ifdef TRACE 
  trace(1,"releasePixmaps");
#endif//TRACE

 if (backgroundPix.pixmap)  XFreePixmap (display, backgroundPix.pixmap); 
 if (maxBoxPix.pixmap)      XFreePixmap (display, maxBoxPix.pixmap); 
 if (stickyBoxPix.pixmap)   XFreePixmap (display, stickyBoxPix.pixmap); 
 if (unStickyBoxPix.pixmap) XFreePixmap (display, unStickyBoxPix.pixmap); 
 if (closeBoxPix.pixmap)    XFreePixmap (display, closeBoxPix.pixmap); 

#ifdef TRACE 
  trace(-1,"releasePixmaps");
#endif//TRACE
}// end releasePixmaps 
#endif//PIXMAPS
/**************************************************************************/
